function subtract_indiv_drifts, drifts_all, stopcorr, maxnoise=maxnoise, highfnoise=highfnoise, $
         ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, ind_aver=ind_aver, $
         file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
         t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, iter_tot=iter_tot, iter_indiv=iter_indiv, $
         clean=clean, time_1=time_1, t_proc=t_proc, n_lt_indiv=n_lt_indiv, parallel=parallel, $
         galactic=galactic, size_deg=size_deg, cut_start=cut_start, cut_nza_weight=cut_nza_weight


;;; computes individual drifts, interpolates and subtracts them from the signal

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2014 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

crs = string(13b) + string(10b)



if (size_deg lt 0.75 and nscans lt 20 and total(cut_start) gt nscans / 5 and (not keyword_set(galactic)) $
   and iter_indiv lt n_lt_indiv) then corr_trans = 1 else corr_trans = 0

if corr_trans gt 0 then begin
   restore, file_signal
   signal_save = signal
   wcut = where(cut_start gt 0, cw_cut)
   for c = 0, cw_cut - 1 do begin
       subs = where(ind_subscans ge ind_scans(wcut(c)))
       signal(ind_subscans(subs(0)):ind_subscans(subs(0)+1)-1, *) = median(signal)
   endfor
   save, file=file_signal, signal, /xdr
   signal = 0
   restore, file_maps
   map_signal_save = map_signal
   map_weight_save = map_weight
   if keyword_set(map_signal_uncl) then map_uncl_save = map_signal_uncl else map_uncl_save = 0
   map_wdrifts_save = map_weight_drifts
   if keyword_set(clean) then proj_map_clean, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, cut_nza_weight=cut_nza_weight $
      else proj_map, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans, cut_nza_weight=cut_nza_weight
   signal = signal_save
   signal_save = 0
   save, file=file_signal, signal, /xdr
   signal = 0
endif


compute_indiv_drifts, array_drifts, array_npass, w_glitches, nb_glitch, $
   maxnoise=maxnoise, highfnoise=highfnoise, ind_scans=ind_scans, ind_aver=ind_aver, $
   file_scalars=file_scalars, min_nsamp=min_nsamp, detect_gl=detect_gl, $
   t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, time_1=time_1, t_proc=t_proc, $
   iter_indiv=iter_indiv, n_lt_indiv=n_lt_indiv, galactic=galactic


if corr_trans gt 0 then begin
   map_signal = map_signal_save
   map_weight = map_weight_save
   map_signal_uncl = map_uncl_save
   map_weight_drifts = map_wdrifts_save
   save, file=file_maps, map_signal, map_weight, map_weight_drifts, map_error, map_lowfnoise, map_signal_uncl, /xdr
   map_signal_save = 0
   map_weight_save = 0
   map_uncl_save = 0
   map_wdrifts_save = 0
endif


;;; mask glitches and update projection vectors:
if nb_glitch ge 1 then begin
   print, crs + 'removed ' + chain(nb_glitch) + ' sample(s) affected by glitches'
   restore, file_signal
   signal(w_glitches) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0

   prepare_proj_coarse, file_scalars=file_scalars
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   w_glitches = 0

   if (keyword_set(visu) and nb_glitch gt 100) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
      if keyword_set(clean) then map_signal = map_signal_uncl
      proj_map, map_noglitch, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      disp_ima, map_signal - map_noglitch, win=4, title='map of residual glitches', $
         weightmap=map_weight, /trunc_max, noise_level=median(maxnoise)
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 4
      time_1 = systime(/seconds)
   endif
endif



restore, file_times

restore, file_time_coarse


drifts = fltarr(nt, nb)
amp_drifts = fltarr(nscans, nb)



npass_min = 0.


;;; stop correction after 3 iterations if it is sparsely determined:

flag_nza_coarse = interpol(flag_nza, float(time_s), time_save)
stopcorr = 0
if (not keyword_set(n_lt_indiv)) then n_lt_indiv = 0
if iter_indiv ge max([2, n_lt_indiv + 2]) then begin
   wnom = where(flag_nza_coarse gt 0., cnom)
   wdef = where(rebin(flag_nza_coarse, nt_save, nb) gt 0. and array_npass gt npass_min, cdef)
   wb = where(finite(maxnoise) gt 0, cb)
   cdef /= (cb / float(nscans))
   if cdef lt 0.5 * cnom then begin
      print, '              drifts sparsely determined -> correction to be stopped after current iteration.'
      stopcorr = 1
   endif
endif


wdef = where(array_npass gt npass_min)
med = median(array_drifts(wdef))

npass_min = fltarr(nscans)
if (not keyword_set(parallel)) then for i = 0, nscans - 1 do begin
   csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
   wsave = ind_sc_coarse(i) + lindgen(csave)
   tab_npass = array_npass(wsave, *)
   wdef = where(rebin(flag_nza_coarse(wsave), csave, nb) gt 0. and tab_npass gt 0., cdef)
   if cdef ge (iter_indiv + 1) * 5 then npass_min(i) = 0.3 * median(tab_npass(wdef), /even) else npass_min(i) = 0.
endfor


for b = 0L, nb - 1L do for i = 0, nscans - 1 do if finite(maxnoise(i, b)) gt 0. then begin
   time_ref = time_s(ind_scans(i))
   csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
   wsave = ind_sc_coarse(i) + lindgen(csave)
   wz = where((array_npass(wsave, b) le npass_min(i) and flag_nza_coarse(wsave) gt 0. $
      and array_npass(wsave, b) lt 0.65 * smooth(array_npass(wsave, b), 7)) or array_npass(wsave, b) le 0. $
      or finite(array_drifts(wsave, b)) lt 1, cz, complement=wdef, ncomplement=cdef)
   if (cdef ge 3 and (iter_indiv lt n_lt_indiv or cdef ge 15)) then begin
      if cz ge 1 then array_drifts(wsave(wz), b) = 0.
      tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
      tab_drifts = [array_drifts(wsave(0), b), array_drifts(wsave, b), array_drifts(wsave(csave-1), b)]
      subs_drifts = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
      drifts(ind_scans(i):ind_scans(i+1)-1, b) = subs_drifts
      w = where(flag_nza(ind_scans(i):ind_scans(i+1)-1) gt 0, cw)
      if cw lt (ind_scans(i+1) - ind_scans(i)) / 3. then dev = sqrt(mean((subs_drifts - med)^2.)) $
         else dev = sqrt(mean((subs_drifts(w) - med)^2.))
      amp_drifts(i, b) = 3. * dev
   endif else begin
      array_drifts(wsave, b) = 0.
      drifts(ind_scans(i):ind_scans(i+1)-1, b) = 0.
   endelse
endif


;;; update total drifts:
drifts_all += array_drifts


restore, file_signal


;;; visualize result for reference bolometer:
if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   window, 0, xs=1200
   wdev = where(finite(signal(*, bref)) gt 0., cw)
   while cw lt 10 do begin
      bref += 1
      wdev = where(finite(signal(*, bref)) gt 0., cw)
   endwhile
   med = median(signal(wdev, bref))
   dev = sqrt(mean((signal(wdev, bref) - med)^2.))
   for k = 0, 9 do begin
      wdev = where(abs(signal(*, bref) - med) lt 3. * dev)
      med = median(signal(wdev, bref))
      dev = sqrt(mean((signal(wdev, bref) - med)^2.))
   endfor
   yrange = [-5.*dev, 8.*dev]
   otherplot, time_s, signal(*, bref) - med, charsize=1.5, yrange=yrange, title=title, $
      xtitle='time (s)', ytitle='brightness + offset'
   otherplot, /over, psym=4, time_save, array_drifts(*, bref) - 3. * dev, color=10
   otherplot, /over, time_s, drifts(*, bref) - 3. * dev, color=100
   for i = 0, nscans - 2 do begin
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)-1) + fltarr(2), yrange
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)) + fltarr(2), yrange
   endfor
   if nsubscans le 30 then otherplot, /over, time_s, dev * (13.8 * flag_nza - 10.), color=200
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

   otherplot, time_s, signal(*, bref) - med, charsize=1.5, yrange=yrange, $
      title='zoom', xtitle='time (s)', ytitle='brightness + offset', $
      xrange=time_s([ind_subscans(0), ind_subscans(min([4, nsubscans]))-1])
   otherplot, /over, psym=4, time_save, array_drifts(*, bref) - 3. * dev, color=10
   otherplot, /over, time_s, drifts(*, bref) - 3. * dev, color=100
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif



print, '              drifts amplitude / noise (median, min and max) = ' $
   + chain(median(amp_drifts / maxnoise), 2) + ' ; ' $
   + chain(min(amp_drifts / maxnoise, /nan), 2) + ' to ' + chain(max(amp_drifts / maxnoise, /nan), 2) + crs



;;; update signal:
signal -= drifts
save, file=file_signal, signal, /xdr


;;; visualization of drifts:
if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   restore, file_maps
   if keyword_set(clean) then map_signal = map_signal_uncl
   proj_map, map_nodrifts, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
   title = 'map of iteration ' + chain(iter_tot) + ' drifts'
   if keyword_set(detect_gl) then title = title + ' (+ glitches)'
   map_drifts = map_signal - map_nodrifts
   disp_ima, map_drifts, win=3, title=title, weightmap=map_weight, noise_level=median(maxnoise)
   map_ratio = fltarr(dimx, dimy)
   w = where(map_signal gt median(maxnoise))
   map_ratio(w) = abs(map_drifts(w)) / map_signal(w)
   disp_ima, map_ratio, win=4, title='ratio map of iteration ' + chain(iter_tot) + ' |drifts| / signal', $
      weightmap=map_weight, noise_level=median(maxnoise), max_map=min([max(map_ratio), 0.15])
   if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif


return, amp_drifts

end
